#ifndef __HTTPS__CLIENT_H__
#define __HTTPS__CLIENT_H__

#include "../../std.h"

#include <globus_io.h>

#include "../../misc/condition.h"
#include "../../misc/url_options.h"


typedef enum {
  HTTP_OK = 200,
  HTTP_NOT_IMPLEMENTED = 501,
  HTTP_NOT_ALLOWED = 403,
  HTTP_NOT_FOUND = 404,
  HTTP_ERROR = 500,
  HTTP_FAILURE = -1
} HTTP_Error;

class HTTP_ResponseHeader {
 private:
  bool keep_alive;

  bool content_length_passed; // Content-Length
  unsigned long long int content_length;

  bool content_range_passed;  // Content-Range
  unsigned long long int content_size;
  unsigned long long int content_start;
  unsigned long long int content_end;
 public:
  HTTP_ResponseHeader(bool alive = true);
  void reset(bool alive = true);
  bool set(const char* name,const char* value);
  bool haveContentRange(void) const { return content_range_passed; };
  bool haveContentLength(void) const { return content_length_passed; };
  unsigned long long int ContentLength(void) const {
    if(content_length_passed) return content_length;
    if(content_range_passed) return (content_end-content_start+1);
    return 0;
  };
  unsigned long long int ContentStart(void) const {
    if(content_range_passed) return content_start;
    return 0;
  };
  unsigned long long int ContentEnd(void) const {
    if(content_range_passed) return content_end;
    if(content_length_passed && content_length) return content_length-1;
    return 0;
  };
  unsigned long long int ContentSize(void) const {
    if(content_range_passed) return content_size;
    if(content_length_passed) return content_length;
    return 0;
  };
  bool KeepAlive(void) const { return keep_alive; };
};

class HTTP_ClientSOAP;

class HTTP_Client {
 friend class HTTP_ClientSOAP;
 private:
  globus_io_handle_t s;
  globus_io_attr_t attr;
  globus_io_secure_authorization_data_t auth;
  URL base_url;
  int timeout;
  Condition<int> cond;
  Condition<int> cond_read;
  Condition<int> cond_write;
  bool valid;
  bool connected;
  char answer_buf[256];
  unsigned int answer_size;
  unsigned int answer_count;
  int answer_code;
  string answer_reason;
  HTTP_ResponseHeader fields;
  static void general_callback(void *arg,globus_io_handle_t *handle,globus_result_t  result);
#ifdef OLD_GLOBUS_IO
  static globus_bool_t authorization_callback(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t* context_handle);
#else
  static globus_bool_t authorization_callback(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t context_handle);
#endif
  static void write_callback(void *arg,globus_io_handle_t *handle,globus_result_t result,globus_byte_t *buf,globus_size_t nbytes);
  static void read_callback(void *arg,globus_io_handle_t *handle,globus_result_t result,globus_byte_t *buf,globus_size_t nbytes);
  int parse_response(void);
  void clear_input(void);
  int analyze_response_line(char* line);
  int read_response_header(void);
  int skip_response_entity(void);
  int PUT_header(const char* path,unsigned long long int offset,unsigned long long int size,unsigned long long int fd_size);
  int GET_header(const char* path,unsigned long long int offset,unsigned long long int size);
 public:
  typedef int (*get_callback_t)(unsigned long long offset,unsigned long long size,char* buf,void* arg);
  typedef int (*put_callback_t)(unsigned long long offset,unsigned long long *size,char* buf);
  HTTP_Client(const char* base);
  ~HTTP_Client(void);
  operator bool(void) { return valid; };
  int connect(void);
  int disconnect(void);
  int PUT(const char* path,unsigned long long int offset,unsigned long long int size,const unsigned char* buf,unsigned long long int fd_size);
  int GET(const char* path,unsigned long long int offset,unsigned long long int size,get_callback_t callback,void* arg);
  bool keep_alive(void) const { return fields.KeepAlive(); };
  unsigned long long int size(void) const { return fields.ContentSize(); };
};

class HTTP_ClientSOAP: public HTTP_Client {
 private:
  struct soap *soap;
  static int local_fsend(struct soap *sp, const char* buf, size_t l);
  static size_t local_frecv(struct soap* sp, char* buf, size_t l);
  static int local_fopen(struct soap* sp, const char* endpoint, const char* host, int port);
  static int local_fclose(struct soap* sp);
 public:
  HTTP_ClientSOAP(const char* base,struct soap *sp);
  ~HTTP_ClientSOAP(void);
};

#endif /*  __HTTPS__CLIENT_H__ */
