#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <vector>
#include <string>
#include <fstream>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "Preferences.h"
#include "DateTime.h"
#include "MdsQuery.h"
#include "Xrsl.h"
#include "Target.h"
#include "JobSubmission.h"

#define __UI_LIBRARY__
#include "ui_uploader.h"

int JobSubmission (vector <Cluster> & clusterlist, const string & oldcluster,
		   const Xrsl & xrsl, const string & joblist,
		   const bool dryrun, const bool dumpxrsl, const int debug) {

  static bool init = false;
  if (!init) {
    srand (time (NULL));
    init = true;
  }

  vector <Target> targetlist;

  for (vector <Cluster>::iterator cli = clusterlist.begin();
       cli != clusterlist.end(); cli++) {

    // skip the old cluster if we are resubmitting
    if (oldcluster == cli->GetName()) continue;

    for (vector <Queue>::iterator qli = cli->queues.begin();
	 qli != cli->queues.end(); qli++) {

      Target target (&*cli, &*qli);

      if (debug) {
	cout << "Cluster: " << target.cluster->GetLongName() << endl;
	cout << "Queue: " << target.queue->GetName() << endl;
      }

      if (target.queue->GetStatus() != "active") {
	if (debug) cout << "Queue rejected due to status: "
			<< target.queue->GetStatus() << endl;
	continue;
      }

      if (target.queue->GetUserFreeCpus() == UNDEFINED) {
	if (debug) cout << "Queue rejected because user not authorized"
			<< endl;
	continue;
      }

      if (target.queue->GetMaxQueuable() != UNDEFINED &&
	  target.queue->GetQueued() >= target.queue->GetMaxQueuable()) {
	if (debug) cout << "Queue rejected because the maximum number "
			<< "of queued jobs is reached" << endl;
	continue;
      }

      string failattr;
      if (xrsl.Test (target, failattr)) return 1;
      if (!target) {
	if (debug) cout << "Queue rejected because it does not match "
			<< "the XRSL specification (" << failattr << ")"
			<< endl;
	continue;
      }

      if (target.CalculateNeededFileSizes()) return 1;

      if (target.GetRemoteSize() == UNDEFINED) {
	if (debug) cout << "Queue rejected because no valid location was "
			<< "found for the inputfiles" << endl;
	continue;
      }

      if (debug > 1) cout << "Total size of remote files: "
			  << target.GetRemoteSize() << " bytes" << endl;
      if (debug > 1) cout << "Total size of local files: "
			  << target.GetLocalSize() << " bytes" << endl;

      if (target.cluster->GetCacheFree() == UNDEFINED) {
	if (debug > 1) cout << "Total size of files: "
			    << target.GetNeededTotalSize()
			    << " bytes" << endl;

	if (target.GetNeededTotalSize() > target.queue->GetUserDiskSpace()) {
	  if (debug) cout << "Queue rejected due to insufficient disk space"
			  << endl;
	  continue;
	}
      }
      else {
	if (debug > 1) cout << "Total size of files to be cached: "
			    << target.GetNeededCacheSize()
			    << " bytes" << endl;
	if (debug > 1) cout << "Total size of files in session directory: "
			    << target.GetNeededSessDirSize()
			    << " bytes" << endl;

	if (target.GetNeededCacheSize() > target.cluster->GetCacheFree()) {
	  if (debug) cout << "Queue rejected due to insufficient cache space"
			  << endl;
	  continue;
	}
	if (target.GetNeededSessDirSize() >
	    target.queue->GetUserDiskSpace()) {
	  if (debug) cout << "Queue rejected due to insufficient disk space"
			  << endl;
	  continue;
	}
      }

      targetlist.push_back (target);
      if (debug) cout << "Queue accepted as possible submission target"
		      << endl;
    }
  }

  bool submitted = false;
  while (!submitted) {

    int freecpus = 0;
    long long int minremotesize = UNDEFINED;
    long long int minlocalsize = UNDEFINED;

    for (vector <Target>::iterator vti = targetlist.begin();
	 vti != targetlist.end(); vti++) {

      if (vti->queue->GetMaxRunning() != UNDEFINED &&
	  vti->queue->GetRunning() >= vti->queue->GetMaxRunning()) continue;

      int time;
      if (vti->xrsl.GetCpuTime (&time)) return 1;
      if (time == UNDEFINED) time = vti->queue->GetDefaultCpuTime();
      int userfreecpus = vti->queue->GetUserFreeCpus (time);

      int count;
      if (vti->xrsl.GetCount (&count)) return 1;
      if (count == UNDEFINED) count = 1;

      if (userfreecpus < count) continue;

      if (minremotesize == UNDEFINED ||
	  vti->GetRemoteSize() < minremotesize) {
	minremotesize = vti->GetRemoteSize();
	minlocalsize = vti->GetLocalSize();
	freecpus = userfreecpus;
      }
      else if (vti->GetRemoteSize() == minremotesize) {
	if (vti->GetLocalSize() < minlocalsize) {
	  minlocalsize = vti->GetLocalSize();
	  freecpus = userfreecpus;
	}
	else
	  freecpus += userfreecpus;
      }
    }

    vector <Target>::iterator target = targetlist.end();

    if (freecpus > 0) {
      float rn = rand() / (RAND_MAX + 1.0) * freecpus;
      int counter = 0;
      bool found = false;

      for (vector <Target>::iterator vti = targetlist.begin();
	   !found && (vti != targetlist.end()); vti++) {

	if (vti->queue->GetMaxRunning() != UNDEFINED &&
	    vti->queue->GetRunning() >= vti->queue->GetMaxRunning())
	  continue;

	int time;
	if (vti->xrsl.GetCpuTime (&time)) return 1;
	if (time == UNDEFINED) time = vti->queue->GetDefaultCpuTime();
	int userfreecpus = vti->queue->GetUserFreeCpus (time);

	int count;
	if (vti->xrsl.GetCount (&count)) return 1;
	if (count == UNDEFINED) count = 1;

	if (userfreecpus < count) continue;

	if (vti->GetRemoteSize() == minremotesize &&
	    vti->GetLocalSize() == minlocalsize) {
	  counter += userfreecpus;
	  if (rn < counter) {
	    target = vti;
	    found = true;
	  }
	}
      }
    }

    else {
      for (vector <Target>::iterator vti = targetlist.begin();
	   vti != targetlist.end(); vti++) {

	int count;
	if (vti->xrsl.GetCount (&count)) return 1;
	if (count == UNDEFINED) count = 1;

	if (vti->GetTotalCpus() < count) continue;

	if (target == targetlist.end() ||
	    (float) vti->GetQueued() / vti->GetTotalCpus() <
	    (float) target->GetQueued() / target->GetTotalCpus())
	  target = vti;
      }
    }

    if (target == targetlist.end()) {
      cerr << "No cluster found" << endl;
      return 1;
    }

    if (debug) {
      cout << target->cluster->GetLongName() << " selected" << endl;
      cout << "queue " << target->queue->GetName() << " selected" << endl;
    }

    target->xrsl.AddSimpleRelation ("queue", target->queue->GetName());
    target->xrsl.AddSimpleRelation ("action", "request");
    target->xrsl.AddSimpleRelation ("savestate", "yes");
    target->xrsl.AddSimpleRelation ("clientxrsl", xrsl.str());

    if (dryrun) target->xrsl.AddSimpleRelation ("dryrun", "dryrun");

    if ((target->xrsl.FixJoin()) ||
	(target->xrsl.FixInOut ("executable", "inputfiles")) ||
	(target->xrsl.FixInOut ("stdin", "inputfiles")) ||
	(target->xrsl.FixInOut ("stdout", "outputfiles")) ||
	(target->xrsl.FixInOut ("stderr", "outputfiles")) ||
	(target->xrsl.FixInOut ("gmlog", "outputfiles")) ||
	(target->xrsl.FixExec()) ||
	(target->xrsl.FixStdin()) ||
	(target->xrsl.FixTime ("starttime")) ||
	(target->xrsl.FixPeriod ("cputime")) ||
	(target->xrsl.FixPeriod ("lifetime"))) return 1;

    char hostname[1024];
    globus_libc_gethostname (hostname, 1024);
    target->xrsl.AddSimpleRelation ("hostname", hostname);

    string jobname;
    if (target->xrsl.GetJobName (jobname)) return 1;

    vector <string> filenames;
    if (target->xrsl.PrepareUpload (filenames)) return 1;

    if ((target->xrsl.Collect (*target, "runtimeenvironment")) ||
	(target->xrsl.Collect (*target, "middleware")) ||
	(target->xrsl.Collect (*target, "opsys"))) return 1;

    if (debug > 1) {
      cout << "Parsed output XRSL:" << endl;
      target->xrsl.Print();
    }

    // dump xrsl

    if (dumpxrsl) {
      cout << target->xrsl.str() << endl;
      submitted = true;
      continue;
    }

    // submit the jobs

    char * shortjobid = NULL;

    int err = ui_uploader (target->cluster->GetContact().c_str(),
			   target->xrsl.str().c_str(), &shortjobid,
			   target->cluster->GetContact().c_str(),
			   RSL_ACTION_REQUEST, filenames, debug);

    if (err || !shortjobid) {
      cout << "Job submission failed to queue " << target->queue->GetName()
	   << " at " << target->cluster->GetLongName() << endl;

      if (shortjobid) {
	cout << "Deleting partially submitted job" << endl;
	vector <string> filenames;
	int err = ui_uploader (target->cluster->GetContact().c_str(), NULL,
			       &shortjobid, NULL, RSL_ACTION_CLEAN,
			       filenames, debug);
	free (shortjobid);
      }

      cout << "Trying remaining queues" << endl;
      targetlist.erase (target);
      continue;
    }

    if (target->Accept()) return 1;

    string jobid = target->cluster->GetContact() + '/' + shortjobid;

    free (shortjobid);

    string oldfilename = GetEnv ("HOME");
    oldfilename.append ("/.ngjobs");
    ifstream oldngjobs (oldfilename.c_str());
    string newfilename = oldfilename + ".tmp";
    ofstream newngjobs (newfilename.c_str());
    string jobidname;
    bool written = false;
    while (getline (oldngjobs, jobidname)) {
      int pos = jobidname.find('#');
      string jobname_ = jobidname.substr (pos + 1);
      if (!written && jobname_ > jobname) {
	newngjobs << jobid << '#' << jobname << endl;
	written = true;
      }
      newngjobs << jobidname << endl;
    }
    if (!written) newngjobs << jobid << '#' << jobname << endl;
    oldngjobs.close();
    newngjobs.close();
    remove (oldfilename.c_str());
    rename (newfilename.c_str(), oldfilename.c_str());

    string histfilename = GetEnv ("HOME");
    histfilename.append ("/.nghistory");
    ofstream nghist (histfilename.c_str(), ios::app);
    nghist << TimeStamp() << "  " << jobid << endl;
    nghist.close();

    if (!joblist.empty()) {
      ofstream joblist (joblist.c_str(), ios::app);
      joblist << jobid << endl;
      joblist.close();
    }

    cout << "Job submitted with jobid " << jobid << endl;

    submitted = true;
  }
}
